/*
 * Copyright (c) 2008 Nick Kossifidis <mickflemm@gmail.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#define	_ATH5K_ANI

#include "ath5k.h"
#include "reg.h"
#include "debug.h"
#include "base.h"

/**********************\
* Set ANI params on HW *
\**********************/

static void ath5k_hw_ani_enable_ofdm_weak_sig(struct ath5k_hw *ah, u8 on){
	struct ath5k_ani_state *ani_state = ah->ah_ani_state;

	if (on) {
		/* Set high thresholds to normal values */
		ath5k_hw_reg_write(ah,
				AR5K_REG_SM(AR5K_ANI_OFDM_WEAK_SIG_M2_COUNT_HIGH_ON,
					AR5K_PHY_WEAK_OFDM_HIGH_THR_M2_COUNT)|
				AR5K_REG_SM(AR5K_ANI_OFDM_WEAK_SIG_M1_HIGH_ON,
					AR5K_PHY_WEAK_OFDM_HIGH_THR_M1)|
				AR5K_REG_SM(AR5K_ANI_OFDM_WEAK_SIG_M2_HIGN_ON,
					AR5K_PHY_WEAK_OFDM_HIGH_THR_M2), 
				AR5K_PHY_WEAK_OFDM_HIGH_THR);

		/* Set low thresholds to normal values */
		ath5k_hw_reg_write(ah,
				AR5K_REG_SM(AR5K_ANI_OFDM_WEAK_SIG_M2_COUNT_LOW_ON,
					AR5K_PHY_WEAK_OFDM_LOW_THR_M2_COUNT)|
				AR5K_REG_SM(AR5K_ANI_OFDM_WEAK_SIG_M1_LOW_ON,
					AR5K_PHY_WEAK_OFDM_LOW_THR_M1)|
				AR5K_REG_SM(AR5K_ANI_OFDM_WEAK_SIG_M2_LOW_ON,
					AR5K_PHY_WEAK_OFDM_LOW_THR_M2), 
				AR5K_PHY_WEAK_OFDM_LOW_THR);

		/* Enable self correlator */
		AR5K_REG_ENABLE_BITS(ah, AR5K_PHY_WEAK_OFDM_LOW_THR,
				AR5K_PHY_WEAK_OFDM_LOW_THR_SELFCOR_EN);

		set_bit(ATH5K_ANI_OFDM_WEAK_SIG_DET_ACTIVE, ani_state->state);
	} else {
		/* Set high thresholds to render detection impossible */
		ath5k_hw_reg_write(ah,
				AR5K_REG_SM(AR5K_ANI_OFDM_WEAK_SIG_M2_COUNT_HIGH_ON,
				AR5K_PHY_WEAK_OFDM_HIGH_THR_M2_COUNT)|
				AR5K_REG_SM(AR5K_ANI_OFDM_WEAK_SIG_M1_HIGH_ON,
				AR5K_PHY_WEAK_OFDM_HIGH_THR_M1)|
				AR5K_REG_SM(AR5K_ANI_OFDM_WEAK_SIG_M2_HIGN_ON,
				AR5K_PHY_WEAK_OFDM_HIGH_THR_M2), 
				AR5K_PHY_WEAK_OFDM_HIGH_THR);

		/* Set low thresholds to render detection impossible */
		ath5k_hw_reg_write(ah,
				AR5K_REG_SM(AR5K_ANI_OFDM_WEAK_SIG_M2_COUNT_LOW_OFF,
				AR5K_PHY_WEAK_OFDM_LOW_THR_M2_COUNT)|
				AR5K_REG_SM(AR5K_ANI_OFDM_WEAK_SIG_M1_OFF,
				AR5K_PHY_WEAK_OFDM_LOW_THR_M1)|
				AR5K_REG_SM(AR5K_ANI_OFDM_WEAK_SIG_M2_OFF,
				AR5K_PHY_WEAK_OFDM_LOW_THR_M2), 
				AR5K_PHY_WEAK_OFDM_LOW_THR);

		/* Disable self correlator */
		AR5K_REG_DISABLE_BITS(ah, AR5K_PHY_WEAK_OFDM_LOW_THR,
				AR5K_PHY_WEAK_OFDM_LOW_THR_SELFCOR_EN);

		clear_bit(ATH5K_ANI_OFDM_WEAK_SIG_DET_ACTIVE, ani_state->state);
	}

	return;
}

static int ath5k_hw_ani_set_ssci_immunity_lvl(struct ath5k_hw *ah, u8 lvl){
	struct ath5k_ani_state *ani_state = ah->ah_ani_state;
	struct ath5k_softc *sc = ah->ah_sc;

	/* Boundary check for immunity level */
	if (lvl > AR5K_ANI_MAX_SSCI_IMMUNITY_LEVEL) {
		ATH5K_ERR(sc, "immunity level out of bounds\n");
		return -EINVAL;
	}

	/* Set coarse_high and coarse_low */
	ath5k_hw_reg_write(ah, 
		AR5K_REG_SM(coarse_high[lvl], AR5K_PHY_AGCCOARSE_HI) |
		AR5K_REG_SM(coarse_low[lvl], AR5K_PHY_AGCCOARSE_LO),
		AR5K_PHY_AGCCOARSE);

	/* Set total desired size and fir_pwr */
	ath5k_hw_reg_write(ah,
			AR5K_REG_SM(total_size[lvl], AR5K_PHY_DESIRED_SIZE_TOT),
			AR5K_PHY_DESIRED_SIZE);

	ath5k_hw_reg_write(ah,
			AR5K_REG_SM(fir_pwr[lvl], AR5K_PHY_SIG_FIRPWR),
			AR5K_PHY_SIG);

	ani_state->ssci_immunity_level = lvl;

	return 0;
}

static void ath5k_hw_ani_apply_parameters(struct ath5k_hw *ah){
	struct ath5k_ani_state *ani_state = ah->ah_ani_state;

	ath5k_hw_ani_set_ssci_immunity_lvl(ah,
			ani_state->ssci_immunity_level);

	ath5k_hw_reg_write(ah,
			AR5K_REG_SM(ani_state->ofdm_cyc_rssi_thr << 1,
			AR5K_PHY_OFDM_SELFCORR_CYPWR_THR1),
			AR5K_PHY_OFDM_SELFCORR);

	ath5k_hw_reg_write(ah, ani_state->cck_weak_sig_thr << 1
			& AR5K_PHY_CCK_CROSSCORR_WEAK_SIG_THR,
			AR5K_PHY_CCK_CROSSCORR);

	ath5k_hw_reg_write(ah,
			AR5K_REG_SM(ani_state->firstep << 1,
			AR5K_PHY_SIG_FIRSTEP),
			AR5K_PHY_SIG);

	return;
}

/********************\
* ANI Initialization *
\********************/

static void ath5k_hw_ani_set_min_immunity(struct ath5k_hw *ah){
	struct ath5k_ani_state *ani_state = ah->ah_ani_state;
	ani_state->ssci_immunity_level = ATH5K_SSCI_IMMUNITY_LEVEL_MIN;
	ani_state->ofdm_cyc_rssi_thr = ATH5K_OFDM_CYC_RSSI_THR_MAX;
	ani_state->cck_weak_sig_thr = ATH5K_CCK_WEAK_SIG_THR_MIN;
	ani_state->firstep = ATH5K_FIRSTEP_MIN;
	ath5k_hw_ani_apply_parameters(ah);
}

static void ath5k_hw_ani_set_max_immunity(struct ath5k_hw *ah){
	struct ath5k_ani_state *ani_state = ah->ah_ani_state;
	ani_state->ssci_immunity_level = ATH5K_SSCI_IMMUNITY_LEVEL_MAX;
	ani_state->ofdm_cyc_rssi_thr = ATH5K_OFDM_CYC_RSSI_THR_MIN;
	ani_state->cck_weak_sig_thr = ATH5K_CCK_WEAK_SIG_THR_MAX;
	ani_state->firstep = ATH5K_FIRSTEP_MAX;
	ath5k_hw_ani_apply_parameters(ah);
}

static void ath5k_hw_ani_reset_state(struct ath5k_ani_state *ani_state){
	clear_bit(ATH5K_ANI_POLL_RUNNING, ani_state->state);
	clear_bit(ATH5K_ANI_FALSE_DETECT_ABOVE_HIGH, ani_state->state);
	clear_bit(ATH5K_ANI_FALSE_DETECT_ABOVE_LOW, ani_state->state);
	clear_bit(ATH5K_ANI_SSCI_POLL_DONE, ani_state->state);
	clear_bit(ATH5K_ANI_OFDM_WEAK_SIG_DET_ACTIVE, ani_state->state);
	clear_bit(ATH5K_ANI_OFDM_WEAK_SIG_POLL_DONE, ani_state->state);
	clear_bit(ATH5K_ANI_CCK_WEAK_SIG_POLL_DONE, ani_state->state);
}

int ath5k_hw_ani_attach(struct ath5k_hw *ah){
	struct ath5k_ani_state *ani_state = ah->ah_ani_state;
	struct ath5k_ani_stats *ani_stats = ah->ah_ani_stats;
	struct ath5k_softc *sc = ah->ah_sc;

	/* Allocate needed structs */
	ani_state = kzalloc(sizeof(struct ath5k_ani_state), GFP_KERNEL);
	if (ani_state == NULL) {
		ATH5K_ERR(sc, "out of memory, ANI disabled\n");
		return -ENOMEM;
	}

	ani_stats = kzalloc(sizeof(struct ath5k_ani_state), GFP_KERNEL);
	if (ani_state == NULL) {
		ATH5K_ERR(sc, "out of memory, ANI disabled\n");
		return -ENOMEM;
	}

	/* And initialize them */
	/* Initial ANI state -> minimum immunity / maximum sensitivity */
	ath5k_hw_ani_set_min_immunity(ah);
	ath5k_hw_ani_enable_ofdm_weak_sig(ah, 1);

	/* Initialize stats */
	ani_stats->ofdm_false_detect_high = ATH5K_ANI_OFDM_FALSE_DETECT_HIGH;
	ani_stats->ofdm_false_detect_low = ATH5K_ANI_OFDM_FALSE_DETECT_LOW;
	ani_stats->cck_false_detect_high = ATH5K_ANI_CCK_FALSE_DETECT_HIGH;
	ani_stats->cck_false_detect_low = ATH5K_ANI_CCK_FALSE_DETECT_LOW;
	ani_stats->max_measurement_duration = ATH5K_ANI_MAX_MEASUREMENT_DURATION;

	set_bit(ATH5K_ANI_RUNNING, ani_state->state);

	return 0;
}

void ath5k_hw_ani_detach(struct ath5k_hw *ah){
	ath5k_hw_ani_set_max_immunity(ah);

	ath5k_hw_set_rx_filter(ah, ath5k_hw_get_rx_filter(ah) &
					~AR5K_RX_FILTER_PHYERR);
	kfree(ah->ah_ani_state);
	kfree(ah->ah_ani_stats);
}

void ath5k_hw_ani_reset(struct ath5k_hw *ah){
	struct ath5k_ani_state *ani_state = ah->ah_ani_state;
	struct ath5k_ani_stats *ani_stats = ah->ah_ani_stats;
	ath5k_hw_ani_reset_state(ani_state);
	ath5k_hw_ani_set_min_immunity(ah);
	ath5k_hw_ani_enable_ofdm_weak_sig(ah, 1);
	ani_stats->ofdm_false_detect_high = ATH5K_ANI_OFDM_FALSE_DETECT_HIGH;
	ani_stats->ofdm_false_detect_low = ATH5K_ANI_OFDM_FALSE_DETECT_LOW;
	ani_stats->cck_false_detect_high = ATH5K_ANI_CCK_FALSE_DETECT_HIGH;
	ani_stats->cck_false_detect_low = ATH5K_ANI_CCK_FALSE_DETECT_LOW;
	ani_stats->max_measurement_duration = ATH5K_ANI_MAX_MEASUREMENT_DURATION;
	set_bit(ATH5K_ANI_RUNNING, ani_state->state);
	ath5k_hw_set_rx_filter(ah, ath5k_hw_get_rx_filter(ah) |
					AR5K_RX_FILTER_PHYERR);
}

/***********************\
* ANI Polling functions *
\***********************/

static void ath5k_hw_ani_poll_ssci(struct ath5k_hw *ah){
	struct ath5k_ani_state *ani_state = ah->ah_ani_state;
	struct ath5k_ani_stats *ani_stats = ah->ah_ani_stats;

	set_bit(ATH5K_ANI_POLL_RUNNING, ani_state->state);

	/* SSCI immunity can't help more, if we are not at max immunity level
	 * increase immunity level and finish ssci poll */
	if ((ani_stats->ofdm_false_detects >= ani_stats->prev_ofdm_false_detects) ||
		(ani_stats->cck_false_detects >= ani_stats->prev_cck_false_detects)) {
		if (ani_state->ssci_immunity_level != AR5K_ANI_MAX_SSCI_IMMUNITY_LEVEL)
			ani_state->ssci_immunity_level++;
		set_bit(ATH5K_ANI_SSCI_POLL_DONE, ani_state->state);
	/* SSCI immunity worked let's try decreasing it a little more to increase
	 * sensitivity */
	} else
		ani_state->ssci_immunity_level--;

	ath5k_hw_ani_set_ssci_immunity_lvl(ah, ani_state->ssci_immunity_level);

	return;
}

static void ath5k_hw_ani_poll_ofdm_weak_sig(struct ath5k_hw *ah){
	struct ath5k_ani_state *ani_state = ah->ah_ani_state;
	struct ath5k_ani_stats *ani_stats = ah->ah_ani_stats;

	set_bit(ATH5K_ANI_POLL_RUNNING, ani_state->state);

	/* Tweaking of OFDM cyclic rssi threshold can't help more, if we are not
	 * at minimum thr (max immunity) decrease it and finish poll */
	if (ani_stats->ofdm_false_detects >= ani_stats->prev_ofdm_false_detects) {
		if (ani_state->ofdm_cyc_rssi_thr != ATH5K_OFDM_CYC_RSSI_THR_MIN)
			ani_state->ofdm_cyc_rssi_thr--;
		set_bit(ATH5K_ANI_OFDM_WEAK_SIG_POLL_DONE, ani_state->state);
	/* Tweaking of OFDM cyclic rssi threshold worked, let's increase it
	 * a little more to increase sensitivity */
	} else if (ani_state->ofdm_cyc_rssi_thr < ATH5K_OFDM_CYC_RSSI_THR_MAX)
		ani_state->ofdm_cyc_rssi_thr++;
	else
		set_bit(ATH5K_ANI_OFDM_WEAK_SIG_POLL_DONE, ani_state->state);

	ath5k_hw_reg_write(ah,
			AR5K_REG_SM(ani_state->ofdm_cyc_rssi_thr << 1,
			AR5K_PHY_OFDM_SELFCORR_CYPWR_THR1),
			AR5K_PHY_OFDM_SELFCORR);

}

static void ath5k_hw_ani_poll_cck_weak_sig(struct ath5k_hw *ah){
	struct ath5k_ani_state *ani_state = ah->ah_ani_state;
	struct ath5k_ani_stats *ani_stats = ah->ah_ani_stats;

	set_bit(ATH5K_ANI_POLL_RUNNING, ani_state->state);

	/* Tweaking of firstep can't help more, if we are not
	 * at maximum thr (max immunity) increase it and finish poll */
	if (ani_stats->cck_false_detects >= ani_stats->prev_cck_false_detects) {
		if (ani_state->firstep != ATH5K_FIRSTEP_MAX)
			ani_state->firstep++;
		set_bit(ATH5K_ANI_CCK_WEAK_SIG_POLL_DONE, ani_state->state);
	/* Tweaking of firstep worked, let's decrease it
	 * a little more to increase sensitivity */
	} else if (ani_state->firstep > ATH5K_FIRSTEP_MIN)
		ani_state->firstep--;
	else
		set_bit(ATH5K_ANI_CCK_WEAK_SIG_POLL_DONE, ani_state->state);

	/* XXX: Poll cck_weak_sig_thr or it only has 2 values ?
	 * pattent indicates that we poll cck_weak_sig_thr instead
	 * of firstep. */

	ath5k_hw_reg_write(ah,
			AR5K_REG_SM(ani_state->firstep << 1,
			AR5K_PHY_SIG_FIRSTEP),
			AR5K_PHY_SIG);

}

/**********************************\
* Results gathering and evaluation *
\**********************************/

static void ath5k_hw_ani_run(struct ath5k_hw *ah){
	struct ath5k_ani_state *ani_state = ah->ah_ani_state;
	struct ieee80211_channel *cur_chan = &ah->ah_current_channel;

	/* ANI is not running, disable PHY error reporting
	 * and return */
	if (!test_bit(ATH5K_ANI_RUNNING, ani_state->state)){
		AR5K_REG_DISABLE_BITS(ah, AR5K_PHY_ERR_FIL, 
		AR5K_PHY_ERR_FIL_OFDM | AR5K_PHY_ERR_FIL_CCK);
		return;
	}

	/* Check if SSCI is the reason we have so many false detects
	 * set ssci immunity level to max and start poll_ssci */
	if (test_bit(ATH5K_ANI_FALSE_DETECT_ABOVE_HIGH, ani_state->state)
	&& !test_bit(ATH5K_ANI_POLL_RUNNING, ani_state->state)
	&& !test_bit(ATH5K_ANI_SSCI_POLL_DONE, ani_state->state)) {
		/* We are above high trheshold so set ssci immunity
		 * to max */
		ath5k_hw_ani_set_ssci_immunity_lvl(ah,
				ATH5K_SSCI_IMMUNITY_LEVEL_MAX);
		ath5k_hw_ani_poll_ssci(ah);
		return;
	} else if (!test_bit(ATH5K_ANI_POLL_RUNNING, ani_state->state)
	&& !test_bit(ATH5K_ANI_SSCI_POLL_DONE, ani_state->state)){
			ath5k_hw_ani_poll_ssci(ah);
			return;
	}

	/* SSCI Poll still running ? */
	if (test_bit(ATH5K_ANI_POLL_RUNNING, ani_state->state) &&
	!test_bit(ATH5K_ANI_SSCI_POLL_DONE, ani_state->state)) {
		ath5k_hw_ani_poll_ssci(ah);
		return;
	}
	
	/* We are back here after SSCI poll so SSCI is not the (only) reason,
	 * tweak weak signal detection */
	if (cur_chan->hw_value & CHANNEL_OFDM
	&& !test_bit(ATH5K_ANI_OFDM_WEAK_SIG_POLL_DONE, ani_state->state)) {

		/* First disable OFDM weak signal detection and
		 * see if it works */
		if (test_bit(ATH5K_ANI_OFDM_WEAK_SIG_DET_ACTIVE, ani_state->state)
		&& !test_bit(ATH5K_ANI_POLL_RUNNING, ani_state->state)) {
			ath5k_hw_ani_enable_ofdm_weak_sig(ah, 0);
			return;
		}

		/* We are below low threshold so OFDM weak signal detection
		 * was the problem, re-enable OFDM weak signal detection 
		 * and tweak ofdm_cyc_rssi threshold to optimize immunity */
		if (!test_bit(ATH5K_ANI_FALSE_DETECT_ABOVE_LOW, ani_state->state)
		&& !test_bit(ATH5K_ANI_POLL_RUNNING, ani_state->state)
		&& !test_bit(ATH5K_ANI_OFDM_WEAK_SIG_POLL_DONE, ani_state->state)) {
			ath5k_hw_ani_enable_ofdm_weak_sig(ah, 1);
			ath5k_hw_ani_poll_ofdm_weak_sig(ah);
			return;
		}

		/* OFDM weak signal detection poll still running ? */
		if (test_bit(ATH5K_ANI_POLL_RUNNING, ani_state->state) &&
		!test_bit(ATH5K_ANI_OFDM_WEAK_SIG_POLL_DONE,
		ani_state->state)) {
			ath5k_hw_ani_poll_ofdm_weak_sig(ah);
			return;
		}

		/* Still above low threshold if we operate in G mode (we also
		 * have cck frames) start cck weak signal poll, if not start again
		 * with OFDM signal detection disabled */
		if (test_bit(ATH5K_ANI_FALSE_DETECT_ABOVE_LOW, ani_state->state)
		&& !test_bit(ATH5K_ANI_OFDM_WEAK_SIG_DET_ACTIVE, ani_state->state)) {
			if (cur_chan->hw_value & CHANNEL_G) {
				ath5k_hw_ani_poll_cck_weak_sig(ah);
				return;
			} else {
				ath5k_hw_ani_reset_state(ani_state);
				set_bit(ATH5K_ANI_OFDM_WEAK_SIG_DET_ACTIVE,
					ani_state->state);
				return;
			}
		}
	}

	if (cur_chan->hw_value & CHANNEL_CCK
	&& !test_bit(ATH5K_ANI_CCK_WEAK_SIG_POLL_DONE, ani_state->state)) {

		/* First increase CCK weak signal threshold & firstep to max and
		 * see if it works */
		if (test_bit(ATH5K_ANI_FALSE_DETECT_ABOVE_LOW, ani_state->state)
		&& ani_state->cck_weak_sig_thr < ATH5K_CCK_WEAK_SIG_THR_MAX
		&& !test_bit(ATH5K_ANI_POLL_RUNNING, ani_state->state)) {
			ani_state->cck_weak_sig_thr = ATH5K_CCK_WEAK_SIG_THR_MAX;
			ath5k_hw_reg_write(ah, ATH5K_CCK_WEAK_SIG_THR_MAX 
					& AR5K_PHY_CCK_CROSSCORR_WEAK_SIG_THR,
					AR5K_PHY_CCK_CROSSCORR);
			ani_state->firstep = ATH5K_FIRSTEP_MAX;
			ath5k_hw_reg_write(ah,
				AR5K_REG_SM(ani_state->firstep << 1,
				AR5K_PHY_SIG_FIRSTEP),
				AR5K_PHY_SIG);
			return;
		}

		/* We are below low threshold so CCK weak signal detection
		 * was the problem, tweak firstep to optimize
		 * immunity */
		if (!test_bit(ATH5K_ANI_FALSE_DETECT_ABOVE_LOW, ani_state->state)
		&& !test_bit(ATH5K_ANI_POLL_RUNNING, ani_state->state)
		&& !test_bit(ATH5K_ANI_CCK_WEAK_SIG_POLL_DONE, ani_state->state)) {
			ath5k_hw_ani_poll_cck_weak_sig(ah);
			return;
		}

		/* CCK weak signal detection poll still running ? */
		if (test_bit(ATH5K_ANI_POLL_RUNNING, ani_state->state) &&
		!test_bit(ATH5K_ANI_CCK_WEAK_SIG_POLL_DONE,
		ani_state->state)) {
			ath5k_hw_ani_poll_cck_weak_sig(ah);
			return;
		}

		/* Still above low threshold if we operate in G mode (we also
		 * have ofdm frames) re-start ofdm weak signal poll else start
		 * again with CCK weak signal threshold set to max */
		if (test_bit(ATH5K_ANI_FALSE_DETECT_ABOVE_LOW, ani_state->state)
		&& !test_bit(ATH5K_ANI_OFDM_WEAK_SIG_DET_ACTIVE, ani_state->state)) {
			if (cur_chan->hw_value & CHANNEL_G) {
				ath5k_hw_ani_enable_ofdm_weak_sig(ah, 1);
				ath5k_hw_ani_poll_ofdm_weak_sig(ah);
				return;
			} else {
				ath5k_hw_ani_reset_state(ani_state);
				return;
			}
		}
	}


	/* All polls have settled and we remain below low false detect threshold,
	 * so we were successful, increase success_count and if it gets above interval
	 * try to decrease ssci immunity parameters */
	if (!test_bit(ATH5K_ANI_FALSE_DETECT_ABOVE_LOW, ani_state->state)
	&& (ani_state->success_count < ATH5K_ANI_MAX_SUCCESS_COUNT)) {
		ani_state->success_count++;
	} else {
		ath5k_hw_ani_set_ssci_immunity_lvl(ah,
				--ani_state->ssci_immunity_level);
	}

	/* All polls have settled and we remain above low false detect threshold,
	 * zero success_count, set immunity to max and disable phy error reporting
	 * to reduce load */
	if (!test_bit(ATH5K_ANI_FALSE_DETECT_ABOVE_LOW, ani_state->state)) {
		ani_state->success_count = 0;
		ath5k_hw_ani_set_max_immunity(ah);
		clear_bit(ATH5K_ANI_RUNNING, ani_state->state);
		ath5k_hw_set_rx_filter(ah,
				ath5k_hw_get_rx_filter(ah) & ~AR5K_RX_FILTER_PHYERR);
	}

	return;
}

/* Called for every PHY error we get during processing of 
 * RX status descriptor.
 * XXX: Locking of ani_stats ? */
void ath5k_hw_ani_report(struct ath5k_hw *ah, u8 rx_error){
	struct ath5k_ani_state *ani_state = ah->ah_ani_state;
	struct ath5k_ani_stats *ani_stats = ah->ah_ani_stats;
	u32 current_time = (jiffies / HZ) * 1000;
	bool is_false_ofdm = 0;
	bool is_false_cck = 0;

	/* It's the first error we have
	 * for this measurement period
	 * mark measurement_start */
	if (!ani_stats->measurement_start) {
		ani_stats->measurement_start = current_time;
		ani_stats->ofdm_false_detects = 0;
		ani_stats->cck_false_detects = 0;
	}

	if (AR5K_REG_MS(rx_error, AR5K_RX_DESC_ERROR1_PHY_ERROR_CODE)
	== AR5K_DESC_RX_PHY_ERROR_OFDM_TIMING)
		is_false_ofdm = 1;
	else if (AR5K_REG_MS(rx_error, AR5K_RX_DESC_ERROR1_PHY_ERROR_CODE)
	== AR5K_DESC_RX_PHY_ERROR_CCK_TIMING)
		is_false_cck = 1;
	else
		return;

	if (ani_stats->measurement_start) {
		/* Measurement duration is not over, update counters
		 * and keep counting */
		if (current_time - ani_stats->measurement_start
		<= ani_stats->max_measurement_duration) {
			if (is_false_ofdm)
				ani_stats->ofdm_false_detects++;
			else if (is_false_cck)
				ani_stats->cck_false_detects++;
			return ;
		}
		/* Measurement period exceeded, reset counters
		 * and evaluate results. Ignore this phy error
		 * since it doesn't belong to this measurement */
		else {
			ani_stats->measurement_start = 0;
			ani_stats->prev_ofdm_false_detects =
				ani_stats->ofdm_false_detects;
			ani_stats->prev_cck_false_detects =
				ani_stats->cck_false_detects;
			if ((ani_stats->ofdm_false_detects >=
				ani_stats->ofdm_false_detect_high) ||
				(ani_stats->cck_false_detects >=
				ani_stats->cck_false_detect_high))
					set_bit(ATH5K_ANI_FALSE_DETECT_ABOVE_HIGH,
						ani_state->state);
			else if ((ani_stats->ofdm_false_detects >=
				ani_stats->ofdm_false_detect_low) ||
				(ani_stats->cck_false_detects >=
				ani_stats->cck_false_detect_low))
					set_bit(ATH5K_ANI_FALSE_DETECT_ABOVE_LOW,
					ani_state->state);
			ani_stats->ofdm_false_detects = 0;
			ani_stats->cck_false_detects = 0;
			ath5k_hw_ani_run(ah);
		}

	}
}

#undef _ATH5K_ANI
